"""
Work with plugins. A plugin is a python module with some special methods.
The PluginManager is used to load on hold these modules.
"""
import sys, os
import toposort
import logger
import imp

log = logger.logging.getLogger('pluginmanager')


class ExtensionPoint:
    """an extension point keeps a bunch of references to objects,
    can be extension point again. extension points have a hierachical
    naming. a dot '.' goes one level deeper.
    e.g. 'root.sub.subsub' etc.
    
    access to an sub extension point is delegated to the object stored
    for the root name.
    trees can be built by using layered ExtensionPoint's.
    e.g.
    root = ExtensionPoint()
    root['child1'] = ExtensionPoint()
    root['child1.other'] = ExtensionPoint()
    """
    def __init__(self):
        self.extensionPoints = {}
    
    def __getitem__(self, point):
        """get an extension point."""
        if '.' in point:
            point, subpoints = point.split('.',1)
            return self.extensionPoints[point][subpoints]
        else:
            return self.extensionPoints[point]

    def __setitem__(self, point, value):
        """register an extension point"""
        if '.' in point:
            point, subpoints = point.split('.',1)
            self.extensionPoints[point][subpoints] = value
        else:
            self.extensionPoints[point] = value

    def __delitem__(self, point):
        """delete an extension point"""
        if '.' in point:
            point, subpoints = point.split('.',1)
            del self.extensionPoints[point][subpoints]
        else:
            del self.extensionPoints[point]

    def __iter__(self):
        return iter(self.extensionPoints)



INITIALIZED     = 1
STARTED         = 2

class Plugin:
    """A plugin has dependencies to extension points, it provides extension
    points, it can be initilized, started, stopped, destroyed.
    basic reload support is also built in"""
    
    def __init__(self, name, module):
        self.state = 0
        self.name = name
        self.module = module
        self.provides = []
        self.depends = []
        self.scan()

    def scan(self):
        """get the provided and dependent extension points"""
        self.provides = []
        self.depends = []
        if hasattr(self.module, 'provides'):
            self.provides = self.module.provides
        if hasattr(self.module, 'depends'):
            self.depends = self.module.depends
    
    def isInitialized(self):
        return bool(self.state & INITIALIZED)

    def isRunning(self):
        return bool(self.state & STARTED)

    def init(self, *args):
        if not (self.state & INITIALIZED):
            if self._do('init', 'initializing', args):
                self.state |= INITIALIZED
        else:
            log.debug('%r already initilized' % (self.name,))
    
    def start(self, *args):
        if self.state & INITIALIZED:
            if not (self.state & STARTED):
                if self._do('start', 'starting', args):
                    self.state |= STARTED
            else:
                log.debug('%r already started' % (self.name,))
        else:
            log.info('%r is not initialized, not starting' % (
                self.name,
            ))
    
    def stop(self, *args):
        if self.state & STARTED:
            if self._do('stop', 'stopping', args):
                self.state &= ~STARTED
        else:
            log.debug('%r already stopped' % (self.name,))
    
    def destroy(self, *args):
        if self.state & STARTED:
            self.stop(*args)
        if self.state & INITIALIZED:
            if self._do('destroy', 'destroying', args):
                self.state &= ~INITIALIZED
        else:
            log.debug('%r already destroyed' % (self.name,))

    def reload(self, *args):
        log.info("reloading %r plugin..." % self.name)
        try:
            running = self.state & STARTED
            if running:
                if self._do('stop', None, args):
                    self.state &= ~STARTED
            if self._do('destroy', None, args):
                self.state &= ~INITIALIZED
            reload(self.module)
            self.scan()
            if self._do('init', None, args):
                self.state |= INITIALIZED
            if running:
                if self._do('start', 'starting', args):
                    self.state |= STARTED
        except:
            log.exception('Error loading %r' % self.name)
    
    def _do(self, functionName, message, args):
        if message:
            log.info("%s %r plugin..." % (message, self.name))
        if hasattr(self.module, functionName):
            try:
                getattr(self.module, functionName)(*args)
            except:
                if message:
                    log.exception('%s failed for %r' % (message, self.name))
                else:
                    raise
                return False
            else:
                return True
        else:
            return True

    def __str__(self):
        return '%s(depends=%s provides=%s)' % (
            self.name, self.depends, self.provides
        )

    __repr__ = __str__

#these extension points are populated by the PluginManager and do not have
#a corresponding plugin file
BUILTIN_EXTENSIONS = [
    'log',
    'plugins',
    'system',
]

class PluginManager(ExtensionPoint):
    """handle a list of plugins (python modules)"""

    def __init__(self):
        ExtensionPoint.__init__(self)
        self.plugins = []
        self.running = 0
        class PluginExtensionPoint:
            def register(inner_self, point, *args, **kwargs):
                raise ValueError("can't register")
            def remove(inner_self, point):
                raise ValueError("can't remove")
            def __getitem__(inner_self, point):
                for plugin in self.plugins:
                    if plugin.name == point:
                        return plugin.module
                else:
                    raise KeyError("no plugin with that name: %r" % point)
            
        self['plugins'] = PluginExtensionPoint()

    def loadPlugins(self, plugindir):
        """load plugins from the specified directory. new plugins are appended
        to the list."""
        log.info('loading plugins from "%s"' % plugindir)
        path = os.path.abspath(plugindir)
        #first load all plugins
        for name in [filename[:-3] for filename in os.listdir(plugindir)
                                   if filename[-3:].lower() == '.py']:
            #only append plugins which are not in the list
            for plugin in self.plugins:
                if plugin.name == name:    #already loaded something with that name
                    break
            else:
                log.info("loading %r plugin...  " % name)
                try:
                    (file, filename, description) = imp.find_module(name, [path])
                    module = imp.load_module(name, file, filename, description)
                except:
                    log.exception("can't load %r" % name)
                else:
                    p = Plugin(name, module)
                    self.plugins.append(p)
        #create dependency tree using "provides" and "depends"
        nodes = BUILTIN_EXTENSIONS[:] #start with the builtin 'providers'
        relations = []
        #get all provided extension points
        for plugin in self.plugins:
            nodes.extend(plugin.provides)
        #check for dependencies that are not provided
        #remove plugins (and dependencies) that do not get all dependencies satisfied
        for plugin in self.plugins[:]:
            for dep in plugin.depends:
                if dep not in nodes:
                    log.error('Extension point %r is not provided but required for %r plugin' % (dep, plugin.name))
                    self.plugins.remove(plugin)
                    for pro in plugin.provides:
                        if pro in nodes:
                            nodes.remove(pro)
        #build a list of dependencies
        for plugin in self.plugins:
            for pro in plugin.provides:
                for dep in plugin.depends:
                    #only add dependencies to existing points
                    if dep in nodes:
                        relations.append((dep, pro))
                    else:
                        log.error('missing %r, used by %r plugin' % (dep, plugin))
        #sort directed graph
        loadOrder = toposort.sort(nodes, relations)
        
        #create load order list by flattening the directed graph
        #save list in _startOrder
        self._startOrder = []
        allplugins = self.plugins[:]
        for l in loadOrder:
            for pro in l:
                found = False
                for plugin in allplugins:
                    if pro in plugin.provides:
                        if plugin not in self._startOrder:
                            self._startOrder.append(plugin)
                            allplugins.remove(plugin)
                            found = True
                            break
                for plugin in self._startOrder:
                    if pro in plugin.provides:
                        found = True
                        break
                if not found and pro not in BUILTIN_EXTENSIONS:
                    log.warn('No plugin provides %r' % pro)
        if allplugins:
            log.warn("Unused plugins, won't load: %s" % (
                ', '.join([p.name for p in allplugins]
            )))
        #save a reversed list too, for shutdown operations
        self._stopOrder = self._startOrder[:]
        self._stopOrder.reverse()
        #start all plugins in order
        for plugin in self._startOrder:
            if not plugin.isInitialized():
                plugin.init(self)
        if self.running:
            for plugin in self._startOrder:
                plugin.start(self)

    def reload(self, name=None):
        """Reload all or a single plugin.
        With no arg (None) all plugins are reloaded, otherwise only the
        specified one."""
        if name is None:
            self.stop()
            for plugin in self.plugins:
                plugin.reload(self)
            self.start()
        else:
            plugin = self['plugins.' + name]
            plugin.reload(self)
    
    def start(self):
        """start all plugins"""
        for plugin in self._startOrder:
            plugin.start(self)
        self.running = 1
    
    def stop(self):
        """stop all plugins"""
        for plugin in self._stopOrder:
            plugin.stop(self)
        self.running = 0
    
    def destroy(self):
        """destroy all plugins"""
        for plugin in self._stopOrder:
            if plugin.isRunning():
                plugin.stop(self)
        for plugin in self._stopOrder:
            plugin.destroy(self)
        self.running = 0
        
    def __str__(self):
        return "%s: %r" % (
            self.__class__.__name__,
            [plugin.name for plugin in self.plugins]
        )

    __repr__ = __str__

class SystemExtensionPoint(ExtensionPoint):
    """registry of system information"""

if __name__ == '__main__':
    #silly tests
    p = PluginManager()
    print p
    p['test'] = 'sample object'
    p['group'] = ExtensionPoint()
    p['group.1'] = 'sample 2'
    print p
    print p['test']
    print p['group']
    print p['group.1']



