"""
Implement the menubar for the main window.

Extension points are registered as menu point by the following scheme:
  'gui.menu.menuname.menupoint'

callbacks can be registered for each menupoint.
"""

import wx

class MenuPoint:
    def __init__(self, callback, label, description=None):
        if description is None: description = label
        self.callback = callback
        self.label = label
        self.description = description
        self.name = None
        self.id = None
    
    def __repr__(self):
        return "%s(callback=%r, label=%r, description=%r)" % (
            self.__class__.__name__,
            self.callback,
            self.label,
            self.description,
        )

class MenuExtensionsHandler:
    """ExtensionPoint handler that handles two layers of extensions at once.
    first layer is the menu name, second the menu point.
    e.g. if 'menu' is a MenuExtensionsHandler instance:
    'menu.help.about'
    
    new menu points must be registered with the factory that is returned by
    'menu.creator'
    such that registering a menuentry looks like the following:
    manager['menu.menuname.menuentry'] = manager['menu.creator'](action, 'name', 'optional description')
    """
    def __init__(self, gui, menu=None):
        self.gui = gui
        if menu is None: menu = {}
        self.menus = menu

    def __getitem__(self, point):
        if point == 'creator':
            return MenuPoint
        elif '.' not in point and self.menus.has_key(point):
            return self.menus[point]
        else:
            menu, name = point.split('.', 1)
            if self.menus.has_key(menu):
                for entry in self.menus[menu]:
                    if entry.name == point:
                        return entry
            raise KeyError("no such element")

    def __setitem__(self, point, value):
    #~ def registerMenu(self, callback, menu, name, description=None):
        if not isinstance(value, MenuPoint):
            raise ValueError("expected a MenuPoint instance")
        if point.count('.') != 1:
            raise ValueError("Submenus not yet supported or no menu entry given")
        menu, id = point.split('.')    #take ext point name as menu name
        #wrap callbacks so that no GUI specific things get back
        def callwrapper(event = None, callback=value.callback):
            callback(self.manager)
        #new menu?
        if not self.menus.has_key(menu):
            self.menus[menu] = []
            self.gui.menus[menu] = wx.Menu()
            #TODO: menu sort order
            self.gui.mainmenu.Append(self.gui.menus[menu], _(menu.capitalize()))
        #remove old entry if any
        del self[point]
        #register new entry
        value.id = wx.NewId()
        value.name = point
        self.gui.menus[menu].Append(value.id, value.label, value.description)
        wx.EVT_MENU(self.gui, value.id, callwrapper)
        self.menus[menu].append(value)
        self.gui.SetMenuBar(self.gui.mainmenu)

    def __delitem__(self, point):
    #~ def removeMenu(self, point):
        menu, name = point.split('.', 1)
        if self.menus.has_key(menu):
            for entry in self.menus[menu]:
                if entry.name == name:
                    self.gui.menus[menu].Remove(entry.id)
                    self.menus[menu].remove(entry)
                    break

    def __iter__(self):
        return iter(self.menus)
    
# - plugin interface:
depends = ['gui']
provides = ['gui.menu']

def init(manager):
    manager['gui.menu'] = MenuExtensionsHandler(manager['gui.handle'])
    manager['gui.menu'].manager = manager
