"""
Implement the toolbar for the main window.

Extension points are registered as tool bar button point by the following
scheme:
  'gui.toolbar.id'

callbacks can be registered for each tool.
"""

import wx

class Tool:
    def __init__(self, callback, description=None, bmp=None):
        if description is None: description = name
        self.callback = callback
        self.description = description
        self.bmp = bmp
        self.name = None
        self.id = None
        
    def __repr__(self):
        return "%s(callback=%r, description=%r)" % (
            self.__class__.__name__,
            self.callback,
            self.description,
        )

class ToolbarExtensionsHandler:
    """ExtensionPoint handler that handles two layers of extensions at once.
    first layer is the menu name, second the menu point.
    e.g. if 'menu' is a MenuExtensionsHandler instance:
    'menu.help.about'
    
    new menu points must be registered with the factory that is returned by
    'menu.creator'
    such that registering a menuentry looks like the following:
    manager['menu.menuname.menuentry'] = manager['menu.creator'](action, 'name', 'optional description')
    """
    def __init__(self, gui, menu=None):
        self.gui = gui
        self.plugin_toolbar = []

    def __getitem__(self, point):
        if point == 'creator':
            return Tool
        elif '.' not in point:
            for entry in self.plugin_toolbar:
                if entry.name == point:
                    return entry
        raise KeyError("no such element")

    def __setitem__(self, point, value):
        if '.' in point:
            raise ValueError("hierarchies not supported")
        if not isinstance(value, Tool):
            raise ValueError("expecting a Tool instance as value")
        #wrap callbacks so that no GUI specific things get back
        def callwrapper(event = None, callback=value.callback):
            callback(self.manager)
        #new or updated button?
        for entry in self.plugin_toolbar:
            if entry.name == point:
                wx.EVT_TOOL(self.gui, entry.id, callwrapper)
                break
        else:
            value.id = wx.NewId()
            value.name = point
            if value.bmp is None:
                value.bmp = self.manager['utility.getBuiltinImage']("question")
            self.gui.toolbar.AddSimpleTool(value.id, value.bmp, value.name, value.description)
            wx.EVT_TOOL(self.gui, value.id, callwrapper)
            self.plugin_toolbar.append(value)
        #refresh display
        self.gui.toolbar.Realize()   #relayout

    def __delitem__(self, point):
        for entry in self.plugin_toolbar:
            (origname, ID_TB) = entry
            if origname == point:
                self.gui.toolbar.RemoveTool(ID_TB)
                self.plugin_toolbar.remove(entry)
                break

    def __iter__(self):
        return iter(self.plugin_toolbar)


# - plugin interface:
depends = ['gui', 'utility']
provides = ['gui.toolbar']

def init(manager):
    manager['gui.toolbar'] = ToolbarExtensionsHandler(manager['gui.handle'])
    manager['gui.toolbar'].manager = manager
